// 每条蛇定义成一个对象
import { AcGameObject } from "./AcGameObject";
import { Cell } from "./Cell";

// 需要继承自绘制基础类 AcGameObject ,用于蛇的绘制刷新
export class Snake extends AcGameObject {
  // 参数:蛇的信息 info ,当前游戏地图 gamemap
  constructor(info, gamemap) {
    super();

    // 每条蛇的 id 用于区分每条蛇, color 用于定义蛇的颜色
    this.id = info.id;
    this.color = info.color;
    // 调用地图的引用获取一些地图参数,例如:每个格子的边长
    this.gamemap = gamemap;

    // 蛇初始只有一个点(蛇头),初始时只需要定义出蛇头即可。初始坐标为每条蛇的起始位置
    // cells[] 存放蛇的身体, cells[0] 存放蛇头
    this.cells = [new Cell(info.r, info.c)];
    this.next_cell = null; // 下一步的目标位置

    this.speed = 5; // 蛇的速度:每秒走五个格子

    // 定义蛇下一步的指令相关属性
    // -1 表示没有指令, 0、 1、 2、 3 表示上右下左方向
    this.direction = -1;
    // idle 表示静止, move 表示移动, die 表示死亡(状态判断的逻辑在蛇群公共部分 GameMap.js 中定义)
    this.status = "idle";

    // 4方向行方向偏移量
    this.dr = [-1, 0, 1, 0];
    // 4方向列方向偏移量
    this.dc = [0, 1, 0, -1];

    // 当前的回合数:前 10 回合,每次蛇身节数都会 +1, 后面每隔 3 回合蛇身节数 +1
    this.step = 0;

    // 定义允许的误差: 0.01 ,当误差为 0.01 以内时,就认为两个点已经重合
    this.eps = 1e-2;

    // 蛇头的眼睛方向:(默认)左下角蛇初始方向朝上
    this.eye_direction = 0;
    // 右上角蛇初始方向朝下
    if (this.id === 1) this.eye_direction = 2;
    // 蛇眼睛四方向x偏移量
    this.eye_dx = [
      [-1, 1],
      [1, 1],
      [1, -1],
      [-1, -1],
    ];

    // 蛇眼睛四方向y偏移量
    this.eye_dy = [
      [-1, -1],
      [-1, 1],
      [1, 1],
      [1, -1],
    ];
  }

  start() {}

  // 定义方向设置接口
  set_direction(d) {
    // 可以将当前方向 this.direction 变为 d , d 是通过键盘或者其他方式人为赋予的方向值
    this.direction = d;
  }

  // 检测当前回合,蛇尾是否增加
  check_tail_increasing() {
    // 前 10 回合每次都增加,后面每 3 回合增加一节蛇尾
    if (this.step <= 10) return true;
    if (this.step % 3 === 1) return true;
    // 否则 return false
    return false;
  }

  // 更新当前状态,将蛇的状态变为走下一步
  next_step() {
    // 当前的蛇头方向
    const d = this.direction;
    // 下一节蛇身体的坐标计算
    this.next_cell = new Cell(
      this.cells[0].r + this.dr[d],
      this.cells[0].c + this.dc[d]
    );

    // 更新蛇眼睛的方向:就是下一步的蛇头方向
    this.eye_direction = d;

    // 计算完坐标之后,清空方向
    this.direction = -1;
    // 将状态从静止变为移动
    this.status = "move";

    // 增加回合数
    this.step++;

    // 计算新的蛇身体结点
    const k = this.cells.length;
    for (let i = k; i > 0; i--) {
      // 每个身体结点都要往后移动一位,配合头部新生成的一位结点,共同组成一个新的蛇身
      // 这里需要使用 JSON 方法进行深度复制,以产生新的对象避免数据出错
      this.cells[i] = JSON.parse(JSON.stringify(this.cells[i - 1]));
    }

    // 如果下一步操作的目标位置碰撞检测不合法,则蛇直接去世
    if (!this.gamemap.check_valid(this.next_cell)) {
      this.status = "die";
    }
  }

  update_move() {
    // 计算目标方向 dx , dy : 使用目标点的坐标减去当前蛇头的坐标
    const dx = this.next_cell.x - this.cells[0].x;
    const dy = this.next_cell.y - this.cells[0].y;
    // 蛇头移动到下一步目标点之间,目前已经移动过的距离
    const distance = Math.sqrt(dx * dx + dy * dy);

    // 计算当前身体结点是否走到终点(下一步的目标位置): 已经移动到终点,则停止移动;
    if (distance < this.eps) {
      // 把目标点存下来作为新的头
      this.cells[0] = this.next_cell;
      // 下一步之前,将当前的 this.next_cell 清空
      this.next_cell = null;

      // 将状态改为 idle 表示当前停下的状态
      this.status = "idle";

      // 如果蛇不边长,则每次移动时,增加头部的同时,把蛇尾砍掉
      if (!this.check_tail_increasing()) {
        this.cells.pop();
      }
    }
    // 不重合表示尚未移动到下一步,还可以继续移动
    else {
      // 按时间(second)定义移动距离: 每一帧走过的距离 = 速度 * 两帧时间间隔 / 1000(ms)
      const move_distance = (this.speed * this.timedelta) / 1000; // 处以 1000 ,将毫秒单位转换成秒单位
      this.cells[0].x += (move_distance * dx) / distance;
      this.cells[0].y += (move_distance * dy) / distance;

      // 更新蛇尾位置
      if (!this.check_tail_increasing()) {
        const k = this.cells.length;
        // 取出当前的蛇尾
        const tail = this.cells[k - 1];
        // 当前蛇尾的下一个目标位置
        const tail_target = this.cells[k - 2];

        // 把当前蛇尾移动到下一个蛇尾目标位置: 将 tail 移动到 tail_target 位置
        // 求两个位置的横纵坐标差值
        const tail_dx = tail_target.x - tail.x;
        const tail_dy = tail_target.y - tail.y;

        // 移动蛇尾
        tail.x += (move_distance * tail_dx) / distance;
        tail.y += (move_distance * tail_dy) / distance;
      }
    }
  }

  // update 方法每一帧执行一次,每秒钟执行 60 次
  update() {
    // 当蛇处于 move 移动状态时, 使用 update_move 方法更新蛇的移动
    if (this.status === "move") {
      // 每一帧调用 this.update_move()
      this.update_move();
    }

    // 每次更新蛇类对象都重新调用一次渲染
    this.render();
  }

  // render 不来自基类,需要本类自己实现
  render() {
    // 取出单元格长度引用和画布
    const L = this.gamemap.L;
    const ctx = this.gamemap.ctx;

    // 定义填充颜色
    ctx.fillStyle = this.color;
    // 当蛇已经去世了,将其颜色变成惨白色
    if (this.status === "die") {
      ctx.fillStyle = "#ffffff";
    }

    // 蛇的身体不止一节,需要枚举画出所有的肢结。of 遍历 cells 值
    for (const cell of this.cells) {
      // 画成正方形
      // ctx.fillRect(cell.c * L, cell.r * L, L, L);

      // 画成圆形
      ctx.beginPath();
      ctx.arc(cell.x * L, cell.y * L, L * 0.5 * 0.8, 0, 2 * Math.PI);
      // 填充颜色
      ctx.fill();
    }

    // 使得蛇身体丰满一点
    for (let i = 1; i < this.cells.length; i++) {
      const a = this.cells[i - 1],
        b = this.cells[i];
      // 当两个目标点重合时,不用在绘制矩形填充
      if (Math.abs(a.x - b.x) < this.eps && Math.abs(a.y - b.y) < this.eps)
        continue;

      // 定义蛇身结点宽度比例
      const snake_node_width = 0.7;
      // 如果两个目标点在竖方向重合(横坐标一致,纵坐标不重合)时的画法
      if (Math.abs(a.x - b.x) < this.eps) {
        ctx.fillRect(
          (a.x - snake_node_width / 2) * L,
          Math.min(a.y, b.y) * L,
          L * snake_node_width,
          Math.abs(a.y - b.y) * L
        );
      }
      // 横方向的画法
      else {
        ctx.fillRect(
          Math.min(a.x, b.x) * L,
          (a.y - snake_node_width / 2) * L,
          Math.abs(a.x - b.x) * L,
          L * snake_node_width
        );
      }
    }

    // 绘制蛇眼睛
    // 定义眼睛颜色
    ctx.fillStyle = "#000000";
    // 定义蛇眼睛大小
    const snake_eye_size = 0.07 * L;
    // 画眼睛
    for (let i = 0; i < 2; i++) {
      // 眼睛的横纵坐标(乘以 L 是绝对距离)
      const eye_x =
        (this.cells[0].x + this.eye_dx[this.eye_direction][i] * 0.15) * L;
      const eye_y =
        (this.cells[0].y + this.eye_dy[this.eye_direction][i] * 0.15) * L;

      ctx.beginPath();
      ctx.arc(eye_x, eye_y, snake_eye_size, 0, 2 * Math.PI);
      ctx.fill();
    }
  }
}
